import swarm.objectbase.*;
import swarm.defobj.*;
import swarm.space.*;
import swarm.gui.*;
import java.util.*;
import java.util.function.*;

public class Bug extends SwarmObjectImpl {
    public float xPos, yPos;
    LinkedList<Vector2D> trailPos;
    int worldXSize, worldYSize;
    Grid2d world;
    Vector2D velocity;
    Vector2D nextVelocity;
    boolean isAvoiding = false;

    public int bugType; // ނ̃IuWFNgǂ邽߂̃tO
    public int moveType; // ړ@ 0: normal Boid, 1: Couzin Boid, 2-3: our impl
    float minSpeed;
    float maxSpeed;
    float accel;
    float gravityWeight;
    float nearWeight;
    float repulRadian;
    float avoidRadian;
    float orientRadian;
    float searchRadian;
    float searchAngle;
    float stdRotate;
    float maxRotate;
    float maxRotateInAvoiding;
    int trailingNum;
    Random rnd;

    public Bug(Zone aZone, float maxSpeed, float minSpeed, float accel, float repulRadian, float avoidRadian, float orientRadian, float searchRadian, float searchAngle, float gravityWeight, float nearWeight, float stdRotate, float maxRotate, float maxRotateInAvoiding, int trailingNum, int bugType, int moveType) {
        super(aZone);
        this.maxSpeed = maxSpeed;
        this.minSpeed = minSpeed;
        this.accel = accel;
        this.repulRadian = repulRadian;
        this.avoidRadian = avoidRadian;
        this.orientRadian = orientRadian;
        this.searchRadian = searchRadian;
        this.searchAngle = searchAngle;
        this.gravityWeight = gravityWeight;
        this.nearWeight = nearWeight;
        this.stdRotate = stdRotate;
        this.maxRotate = maxRotate;
        this.maxRotateInAvoiding = maxRotateInAvoiding;
        this.trailingNum = trailingNum;
        this.bugType = bugType;
        this.moveType = moveType;
        this.rnd = new Random();
        this.trailPos = new LinkedList<Vector2D>();
    }

    public Bug setWorld(Grid2d w) {
        world = w;
        worldXSize = world.getSizeX();
        worldYSize = world.getSizeY();
        return this;
    }

    public Bug setX$Y(int x, int y) {
        xPos = (float)x;
        yPos = (float)y;
        world.putObject$atX$Y(this, x, y);
        return this;
    }

    public Bug setVelocity(float speed, float angle) {
        velocity = Vector2D.createVectorByNormDeg(speed, angle);
        return this;
    }

    public Bug setWorldSizeX$Y(int xSize, int ySize) {
        worldXSize = xSize;
        worldYSize = ySize;
        return this;
    }

    public Vector2D getVelocity() {
        return this.velocity.getCopyVec();
    }

    public Vector2D getPosition() {
        return Vector2D.createVectorByXY(this.xPos, this.yPos);
    }

    // ԂvZ邪AXV͂Ȃ
    public void calc() {
        this.nextVelocity = getNextVelocity();
    }

    // vZꂽԂɑJڂ
    public void step() {
        this.velocity = this.nextVelocity;
        // ړ
        float newX = adjust(xPos + velocity.getX(), worldXSize);
        float newY = adjust(yPos + velocity.getY(), worldYSize);
        // "bug" Փ˂Ƃ͈ʒuXVȂ
        if (world.getObjectAtX$Y((int)newX, (int)newY) == null ||
        ((int)xPos == (int)newX && (int)yPos == (int)newY) ) {
            world.putObject$atX$Y(null, (int)xPos, (int)yPos);
            this.trailPos.addFirst(this.getPosition());
            if (this.trailPos.size() > trailingNum) {
                this.trailPos.removeLast();
            }
            xPos = newX;
            yPos = newY;
            world.putObject$atX$Y(this, (int)newX, (int)newY);
        }
    }

    private Vector2D getNextVelocity() {
        // ̈ʒu, x
        Vector2D myPosition = this.getPosition();
        float nowSpeed = this.velocity.getNorm();
        float nowAngle = this.velocity.getAngle();
        // old-fashioned Boid
        if (moveType == 0) {
            // ͈͓̓bug擾
            ArrayList<Bug> visibleSameTypeBugs = getVisibleBugs(bugType, Optional.empty(), searchRadian, Optional.empty());
            int bugNum = visibleSameTypeBugs.size();
            // dS̈ʒuvZ
            Vector2D gravityDirection = Vector2D.Zero();
            if (bugNum > 0) {
                gravityDirection = visibleSameTypeBugs.stream()
                .map(b -> b.getPosition())
                .reduce((v1, v2) -> v1.add(v2))
                .get()
                .mul((float) 1.0 / bugNum)
                .sub(myPosition)
                .unit();
            }
            // ̕ςvZ
            Vector2D averageDirection = Vector2D.Zero();
            if (bugNum > 0) {
                averageDirection = visibleSameTypeBugs.stream()
                .map(b -> b.getVelocity())
                .reduce((v1, v2) -> v1.add(v2))
                .get()
                .mul((float) 1.0 / bugNum)
                .unit();
            }
            // ԋ߂bug̈ړvZ
            Vector2D nearestDirection = Vector2D.Zero();
            // ԋ߂bug
            Optional<Bug> nearestBug = getVisibleBugs(0, Optional.empty(), searchRadian, Optional.empty()).stream()
            .min(Comparator.comparing(
                b -> myPosition.getDistance(b.getPosition())));
            // ړ
            nearestDirection = nearestBug
            .map(b -> b.getVelocity())
            .map(v -> v.unit())
            .orElse(Vector2D.Zero());
            // type
            int nearestType = nearestBug
            .map(b -> b.bugType)
            .orElse(-1);

            //  time step ̑xxNǧvZ
            // v_{t+1} = v_t + W_g * (v_{grav} + v_{ave}) + d_{type?} * w_n * v_{near}
            Vector2D newDirection = this.velocity
            .getCopyVec()
            .unit()
            .add(
                gravityDirection
                    .add(averageDirection)
                    .mul(gravityWeight))
            .add(
                nearestDirection
                    .mul(nearWeight)
                    .mul(nearestType == this.bugType ? 1.0f : -1.0f))
            .unit();
            //  time step ̑xxNg̑傫vZ
            // nearest bug ̋ repulRadian ߂Ȃ牓AȂ߂Â
            int accelDir = 0; // Ȃ1, Ȃ-1, Ȃ0
            if (nearestBug.isPresent()) {
                accelDir = 1;
                Vector2D nearestPos = nearestBug.map(b -> b.getPosition()).get();
                accelDir *= sign(
                    nearestPos
                        .getCopyVec()
                        .sub(myPosition)
                        .getInnerProd(newDirection));
                accelDir *= sign(nearestPos.getDistance(myPosition) - repulRadian);
            }
            float newSpeed = nowSpeed * (float) Math.pow(accel, accelDir);
            // Xs[hclipping
            if (newSpeed > maxSpeed) {
                newSpeed = maxSpeed;
            }
            if (newSpeed < minSpeed) {
                newSpeed = minSpeed;
            }
            return newDirection.clip(newSpeed, newSpeed);
        } else if (moveType == 1) { // Couzin's algorithm
            Vector2D newDirection;
            // zone of repulsion
            // 1st priority
            // repulsion͈͓bugJEg
            ArrayList<Bug> repulBugs = getVisibleBugs(bugType, Optional.empty(), repulRadian, Optional.empty());
            newDirection = repulBugs.stream()
                .map(b -> b.getPosition())
                .map(v -> v.sub(myPosition))
                .map(v -> v.unit())
                .reduce((v1, v2) -> v1.add(v2))
                .map(v -> v.mul(-1))
                .orElse(Vector2D.Zero());
            if (newDirection.isZero()) {
                // zone of orientation
                // 2nd priority - 1
                // orientation͈͓bugJEg
                ArrayList<Bug> orientBugs = getVisibleBugs(bugType, Optional.of(repulRadian), orientRadian, Optional.of(searchAngle));
                Vector2D orientDirection = orientBugs.stream()
                    .map(b -> b.getVelocity())
                    .map(v -> v.unit())
                    .reduce((v1, v2) -> v1.add(v2))
                    //.map(v -> v.unit())
                    .orElse(Vector2D.Zero());
                // zone of attraction
                // 2nd priority - 2
                // attraction͈͓bugJEg
                ArrayList<Bug> attractBugs = getVisibleBugs(bugType, Optional.of(orientRadian), searchRadian, Optional.of(searchAngle));
                Vector2D attractDirection = attractBugs.stream()
                    .map(b -> b.getPosition())
                    .map(v -> v.sub(myPosition))
                    .map(v -> v.unit())
                    .reduce((v1, v2) -> v1.add(v2))
                    //.map(v -> v.unit())
                    .orElse(Vector2D.Zero());
                if (orientDirection.isZero() && attractDirection.isZero()) {
                    newDirection = this.getVelocity();
                } else {
                    newDirection = orientDirection.add(attractDirection);
                }
            }
            if (newDirection.isZero()) {
                newDirection = Vector2D.createRandomVec(1.0f, 1.0f);
            }
            Vector2D nowDirection = this.getVelocity().unit();
            float diffAngle = nowDirection.getInnerAngle(newDirection) + (float)rnd.nextGaussian() * stdRotate;
            if (Math.abs(diffAngle) > maxRotate) {
                diffAngle = sign(diffAngle) * maxRotate;
            }
            return Vector2D.createVectorByNormDeg(minSpeed, nowAngle + diffAngle);
        } else { // if (moveType == 2 || moveType == 3) {
            // ߐHҁEߐH
            Vector2D newDirection;
            // zone of avoidance / predation
            // 1st priority
            // ڎ͈͓ɊOGD悵Ĕ / lD悵ďP
            isAvoiding = true;
            ArrayList<Bug> avoidBugs1 = getVisibleBugs(-bugType, Optional.empty(), repulRadian, Optional.of(searchAngle));
            newDirection = avoidBugs1.stream()
                .map(b -> b.getPosition())
                .map(v -> v.sub(myPosition))
                .reduce((v1, v2) -> v1.add(v2))
                .map(v -> v.mul(moveType == 2 ? -1 : 1))
                .orElse(Vector2D.Zero());
            if (newDirection.isZero()) {
                ArrayList<Bug> avoidBugs2 = getVisibleBugs(-bugType, Optional.of(repulRadian), avoidRadian, Optional.of(searchAngle));
                newDirection = avoidBugs2.stream()
                    .map(b -> b.getPosition())
                    .map(v -> v.sub(myPosition))
                    .reduce((v1, v2) -> v1.add(v2))
                    .map(v -> v.mul(moveType == 2 ? -1 : 1))
                    .orElse(Vector2D.Zero());
            }
            if (newDirection.isZero()) {
                isAvoiding = false;
                // zone of repulsion
                // 2nd priority
                // repulsion͈͓bugJEg
                ArrayList<Bug> repulBugs = getVisibleBugs(bugType, Optional.empty(), repulRadian, Optional.empty());
                newDirection = repulBugs.stream()
                    .map(b -> b.getPosition())
                    .map(v -> v.sub(myPosition))
                    .map(v -> v.unit())
                    .reduce((v1, v2) -> v1.add(v2))
                    .map(v -> v.mul(-1))
                    .orElse(Vector2D.Zero());
            }
            if (newDirection.isZero()) {
                // zone of orientation
                // 3rd priority - 1
                ArrayList<Bug> orientBugs = getVisibleBugs(bugType, Optional.of(repulRadian), orientRadian, Optional.of(searchAngle));
                Vector2D orientDirection = orientBugs.stream()
                    .map(b -> b.getVelocity())
                    .map(v -> v.unit())
                    .reduce((v1, v2) -> v1.add(v2))
                    //.map(v -> v.unit())
                    .orElse(Vector2D.Zero());
                // zone of attraction
                // 3rd priority - 2
                ArrayList<Bug> attractBugs = getVisibleBugs(bugType, Optional.of(orientRadian), searchRadian, Optional.of(searchAngle));
                Vector2D attractDirection = attractBugs.stream()
                    .map(b -> b.getPosition())
                    .map(v -> v.sub(myPosition))
                    .map(v -> v.unit())
                    .reduce((v1, v2) -> v1.add(v2))
                    //.map(v -> v.unit())
                    .orElse(Vector2D.Zero());
                if (orientDirection.isZero() && attractDirection.isZero()) {
                    newDirection = this.getVelocity();
                } else {
                    newDirection = orientDirection
                    .add(attractDirection);
                }
            }
            if (newDirection.isZero()) {
                newDirection = Vector2D.createRandomVec(1.0f, 1.0f);
            }
            Vector2D nowDirection = this.getVelocity().unit();
            float diffAngle = nowDirection.getInnerAngle(newDirection) + (float)rnd.nextGaussian() * stdRotate;
            float nowMax = isAvoiding ? maxRotateInAvoiding : maxRotate;
            if (Math.abs(diffAngle) > nowMax) {
                diffAngle = sign(diffAngle) * nowMax;
            }
            float newSpeed = nowSpeed * (float) Math.pow(accel, isAvoiding ? 1 : -1);
            // Xs[hclipping
            if (newSpeed > maxSpeed) {
                newSpeed = maxSpeed;
            }
            if (newSpeed < minSpeed) {
                newSpeed = minSpeed;
            }
            return Vector2D.createVectorByNormDeg(newSpeed, nowAngle + diffAngle);
        }
    }

    // ͈͓bugSĕԂ
    private ArrayList<Bug> getVisibleBugs(int type, Optional<Float> minopt, float max, Optional<Float> perceptionDeg) {
        // ̈ʒu
        Vector2D myPosition = this.getPosition();
        ArrayList<Bug> visibleBugs = new ArrayList<Bug>();
        float min = minopt.orElse(0.0f);
        int xMin = (int)(Math.floor(min));
        int xMax = (int)(Math.ceil(max));
        for (int dirI = -1; dirI <= 1; dirI += 2) {
            for (int i = xMin; i <= xMax; i++) {
                int yMin = (min <= i) ? 0 :
                    (int) Math.floor(Math.sqrt(min*min - i*i));
                int yMax = (max <= i) ? 0 :
                    (int) Math.floor(Math.sqrt(max*max - i*i));
                for (int dirJ = -1; dirJ <= 1; dirJ += 2) {
                    for (int j = yMin; j <= yMax; j++) {
                        if (i == 0 && j == 0) {
                            continue;
                        }
                        int nowX = adjust((int)xPos + i * dirI, worldXSize);
                        int nowY = adjust((int)yPos + j * dirJ, worldYSize);
                        // ̈ʒubugo
                        Bug foundBug = (Bug) world.getObjectAtX$Y(nowX, nowY);
                        if (foundBug == null || foundBug == this) {
                            continue;
                        }
                        // {ɔ͈͓mF(intfloat̍̂)
                        if (foundBug.isCanBeSeenTheBug(this, minopt, Optional.of(max), perceptionDeg)) {
                            // type0̏ꍇ͑SĂbugXgɓ
                            // typȅꍇ͓bugXgɓ
                            // typȅꍇ͂typeȊObugXgɓ
                            if (type == 0 ||
                                (type > 0 && foundBug.bugType == type) ||
                                (type < 0 && foundBug.bugType != -type)) {
                                visibleBugs.add( foundBug );
                            }
                        }
                    }
                }
            }
        }
        return visibleBugs;
    }

    //͈͂OȂWlԂ
    private float adjust(float x, int width) {
      if( x >= 0 ) {
        return x % width;
      }
      else {
        return width + ( x % width );
      }
    }
    private int adjust(int x, int width) {
      if( x >= 0 ) {
        return x % width;
      }
      else {
        return width + ( x % width );
      }
    }
    // Ώۂbugthisuvǂ肷
    private boolean isCanBeSeenTheBug(Bug observer, Optional<Float> min, Optional<Float> max, Optional<Float> perceptionDeg) {
        float angleDiff = observer.getPosition().getInnerAngle(this.getPosition());
        boolean isAngleOK = Math.abs(angleDiff) <= perceptionDeg.orElse(360.0f) / 2.0f;
        return isAngleOK && this.isDistanceWithIn(observer, min, max);
    }
    private boolean isCanBeSeenTheBug(Bug observer, float min, float max, float perceptionDeg) {
        return this.isCanBeSeenTheBug(observer, Optional.of(min), Optional.of(max), Optional.of(perceptionDeg));
    }
    // thisƑΏۂbug̋minȏmaxǂ肷
    private boolean isDistanceWithIn(Bug anotherBug, Optional<Float> min, Optional<Float> max) {
        float dist = this.getPosition().getDistance(anotherBug.getPosition());
        float infty = Float.POSITIVE_INFINITY;
        return min.orElse(-infty) <= dist && dist < max.orElse(infty);
    }
    private boolean isDistanceWithIn(Bug anotherBug, float min, float max) {
        return this.isDistanceWithIn(anotherBug, Optional.of(min), Optional.of(max));
    }
    // _0镄֐
    private int sign(float x) {
        if (x == 0.0f) {
            return 0;
        } else if (x > 0.0f) {
            return 1;
        } else {
            return -1;
        }
    }

    public Object drawSelfOn(Raster r) {
        int color = 1 + 10 * (bugType - 1 + (moveType == 2 && isAvoiding ? 1 : 0));
        r.drawPointX$Y$Color((int)xPos, (int)yPos, (byte)color);
        for (Vector2D lastPos : trailPos) {
            color += 1;
            r.drawPointX$Y$Color((int)lastPos.getX(), (int)lastPos.getY(), (byte)color);
        }
        return this;
    }

    public float Vector2Direction(float vx, float vy) {
        return (float) Math.atan(vy / vx);
    }
}
